# ~ encoding: utf-8  ~
import time
import unittest
from mock import MagicMock, patch


import tbbackoff as tbb


class RateLimitTestCase(unittest.TestCase):

    def setUp(self):

        # Stop the time
        self.time_patcher = patch('tbbackoff.time')
        self.time_mock = self.time_patcher.start()
        self.time_mock.time.return_value = time.time()

        def sleep_mock(i):
            self.time_mock.time.return_value += i

        self.time_mock.sleep.side_effect = sleep_mock

    def tearDown(self):
        self.time_patcher.stop()

    def test_first_request_is_not_ratelimited(self):
        def ratelimited_function():
            pass

        ratelimit = tbb.RateLimit()
        function = ratelimit(ratelimited_function)

        function()

        self.assertFalse(self.time_mock.sleep.called)

    def test_ratelimit_after_bucket_empty(self):
        def ratelimited_function():
            pass

        ratelimit = tbb.RateLimit(5, token_time=0.5)
        function = ratelimit(ratelimited_function)

        # Se ejecuta hasta el borde del ratelimit
        for i in range(0, ratelimit.bucket_fill):
            function()

        # No hubo ratelimit
        self.assertFalse(self.time_mock.sleep.called)

        # Al ejecutar mas, tiene que hacer ratelimit
        function()
        self.assertTrue(self.time_mock.sleep.called)

        # Si pasa el tiempo debería llenarse el bucket
        self.time_mock.time.return_value += 3
        function()

        # El bucket está lleno - 1 token
        self.assertEqual(ratelimit.bucket_fill, 4)

    def test_bucket_is_emptied_on_backpressure(self):
        ''' Si se obtiene un 429 el bucket size se reduce
        '''

        empty_bucket_function = MagicMock()

        ratelimit = tbb.RateLimit(6, token_time=0.5, empty_bucket_callback=empty_bucket_function)

        function_mock = MagicMock()
        function_mock.side_effect = [tbb.RateLimitBackPressure(), 'OK']
        function_mock.__name__ = 'function'
        function = ratelimit(function_mock)

        function()

        # El bucket se vacía
        self.assertTrue(empty_bucket_function.called)

    def test_backoff_on_backpressure(self):

        backoff_callback_fn = MagicMock()

        ratelimit = tbb.RateLimit(6, token_time=1, backoff_callback=backoff_callback_fn)

        pressure_response = [tbb.RateLimitBackPressure()]
        ok_response = ['OK']

        function_mock = MagicMock()
        function_mock.side_effect = pressure_response * 3 + ok_response * 8
        function_mock.__name__ = 'function'
        function = ratelimit(function_mock)

        # Al llamar a la función causa 3 backpressure, se recibe la respuesta
        response = function()
        self.assertEqual(response, 'OK')

        # Y se incrementó el backoff
        self.assertEqual(ratelimit.backoff, 4)
        self.assertEqual(backoff_callback_fn.call_count, 3)

        # A medida que pasa el tiempo con llamadas sin backpressure, el backoff se va reduciendo
        self.time_mock.time.return_value += ratelimit.backoff * 8
        high_backoff = ratelimit.backoff

        for i in range(4):
            function()

        self.assertLess(ratelimit.backoff, high_backoff)

        # El backoff se ejecutó cuatro veces más para achicarlo
        self.assertEqual(backoff_callback_fn.call_count, 7)

        # Hasta que llega a cero
        function()
        self.assertEqual(ratelimit.backoff, 0)

        # Y no se reduce mas
        function()
        self.assertEqual(ratelimit.backoff, 0)

    def test_max_retries(self):
        def bad_function():
            raise tbb.RateLimitBackPressure()

        function = tbb.RateLimit(max_retries=4)(bad_function)

        with self.assertRaises(tbb.MaxRetriesReached):
            function()



# ~ encoding: utf-8  ~

import time
import random
from functools import wraps


class RateLimitBackPressure(Exception):
    pass


class MaxRetriesReached(Exception):
    pass


class RateLimit(object):
    """Ratelimit token bucket con backoff

    Se inicializa con los parámetros de bucket y retries y luego se usa como decorador de las funciones a limitar.

    Las funciones deben levantar la excepción RateLimitBackPressure para indicar que se debe incrementar el backoff.

    Si se llamó a la función demasiadas veces, se levanta la excepción MaxRetriesReached
    """

    def __init__(self,
                 bucket=20,
                 token_time=None,
                 max_retries=10,
                 wait_callback=lambda x: None,
                 empty_bucket_callback=lambda: None,
                 backoff_callback=lambda x, y: None,
                 scale_down_factor=3,
                 max_backoff=60,
                 ):

        self.scale_down_factor = scale_down_factor
        self.bucket_fill = bucket
        self.bucket_size = bucket

        if token_time is None:
            random.seed(time.clock())
            token_time = random.uniform(0.3, 0.8)
        self.token_time = token_time

        self.max_retries = max_retries
        self.max_backoff = max_backoff

        self.last_pressure = 0
        self.backoff = 0
        self.last_tick = time.time()

        self.wait_callback = wait_callback
        self.empty_bucket_callback = empty_bucket_callback
        self.backoff_callback = backoff_callback

    @property
    def under_pressure(self):
        return (time.time() - self.last_pressure) < (self.backoff * self.scale_down_factor)

    def replenish_bucket(self):
        time_delta = time.time() - self.last_tick

        token_time = max(self.token_time, self.backoff)

        for i in range(int(time_delta / token_time)):
            if self.bucket_fill < self.bucket_size:
                self.last_tick = time.time()
                self.bucket_fill += 1

    def sleep_and_notify(self, secs):
        time.sleep(secs)
        self.wait_callback(secs)

    def set_backoff(self, new_value):
        old_value = self.backoff
        if new_value > self.max_backoff:
            new_value = self.max_backoff
        self.backoff = new_value
        self.backoff_callback(new_value, old_value)

    def __call__(self, f):
        @wraps(f)
        def ratelimited_function(*args, **kwargs):
            n = 1
            while n <= self.max_retries:
                self.replenish_bucket()

                if self.bucket_fill == 0:
                    self.sleep_and_notify(self.token_time)
                    self.empty_bucket_callback()
                    continue

                if self.backoff:
                    self.sleep_and_notify(self.backoff)

                self.bucket_fill -= 1

                try:
                    n += 1
                    result = f(*args, **kwargs)

                    if not self.under_pressure:
                        self.set_backoff(max(0, self.backoff - 1))

                    return result
                except RateLimitBackPressure:
                    self.bucket_fill = 0

                    self.last_pressure = time.time()

                    if self.backoff == 0:
                        self.set_backoff(self.token_time)
                    else:
                        self.set_backoff(self.backoff * 2)
                    continue
            raise MaxRetriesReached()

        return ratelimited_function



from setuptools import setup
import os.path

# The text of the README file
README = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Readme.md')).read()

# This call to setup() does all the work
setup(
    name="tbbackoff",
    version="0.0.5",
    description="Implementacion de token bucket con backoff exponencial",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/radiocutfm/tbbackoff",
    author="Lambda Sistemas",
    author_email="desarrollo@fierro.com.ar",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
    ],
    packages=["tbbackoff"],
    include_package_data=True,
    install_requires=[],
    entry_points={
        "console_scripts": [
            "realpython=reader.__main__:main",
        ]
    },
)
